///////////////////////////////////////////////////////////////////////////////////////////////
// romclasses.h

class CRom;
#include "rom.h"



//////////////////////////////////////////////////////////////////////
// Classes
// (New section)
// Start of Object-orientation for Map Editor ROM code. Yay!
//


// Abstract base class for a data entry in the ROM
class CDataEntry
{
protected:
	enum class_info
	{
		ptr_address = -1,
		ptr_size = -1,
		def_data_start = -1,
		def_data_extent = -1
	};

public:
	int index;						// Index of this object in table
	long address;					// Address of this object

	// Note that the index value for an object of a derived class
	// is meaningless until ReadObject has been called, specifying
	// which index the object is to have

public:
	// Pure virtual functions must be overloaded
	virtual void ReadObject(FILE* file, int index, DWORD data = 0) = 0;
	virtual void WriteObject(FILE* file, unsigned long address) = 0;
	virtual void WritePtr(FILE* file, unsigned long address) = 0;

	virtual int GetAddress() = 0;
	virtual int GetSize() = 0;


	// The overloaded form of ReadObject in each derived class
	// should use the given file object to seek to the specified
	// entry (index) in that object class's pointer table, and
	// follow that pointer to find and read the data in any way
	// necessary.
	// Note that this approach does not work with unorganized
	// data like door destinations, as they are not indexed by
	// a pointer table, but rather _by_ pointers on an instance-
	// by-instance basis. Door destinations, of course, are a
	// pain in the ass for a lot of reasons, so special measures
	// must be taken when dealing with them.


	CDataEntry() {};
	virtual ~CDataEntry() {};


};


// Other classes are derived from the DataEntry class to
// provide for specific functionality pertinent to other
// editing functions.
// 

// The theory behind these classes is that, although each
// instance of the class represents a single entry of its
// type, the class itself sort of represents a theoretical
// collection of these entries (philosophically, not
// literally), as their intended use involves being implemented
// as a very large collection of said data entries.

class CEnemyCluster : public CDataEntry
{

//////////////////////////////////////////////////////////////////////
// Default implementation
public:

	enum class_info
	{
		ptr_address = 0x10BA80,			// Address of pointer table
		ptr_size = 4,					// Size of each entry in pointer table
		def_data_start = 0x10BDAC,		// Location of original data
		def_data_extent = 0x10C80D,		// Original extent of data
		num_entries = 203,
	};
	// The last two entries in this enumerator are used to keep
	// track of where the data for this class of entry was stored
	// originally; e.g., the code that stores the data for the
	// entire collection of enemy clusters will check these values
	// (as well as call the GetSize member function (TBI)) to
	// determine if the next entry can be written into that original
	// data extent, or if it will have to be bundled into another
	// collection to be later written into the expanded area.
	// Rather neat, no?

	void WriteObject(FILE*, unsigned long address);
	void WritePtr(FILE*, unsigned long address);
	void ReadObject(FILE* file, int index, DWORD data = 0);

	int GetAddress();
	int GetSize();


///////////////////////////////////////////////////////////////////////
// Class-specific data
public:

	// Logical redundancies
	unsigned char n_pri;	// Number of primary entries
	unsigned char n_sec;	// Number of secondary entries
	bool has_primary;		// Indicates if group has primary entry
	bool has_secondary;		// Indicates if group has secondary entry

	// Entry range data
	unsigned short event;	// Event flag that toggles between primary and secondary
	unsigned char p_pri;	// Encounter rate for primary cluster
	unsigned char p_sec;	// Encounter rate for secondary cluster

	struct
	{
		unsigned char prb;
		unsigned short group;
	} primary[8], secondary[8];


	void AddGroup(int nSubCluster, int nFreq, int nGroup);
	void RemGroup(int nSubCluster, int nItem);
};



//////////////////////////////////
// CEnemyBattleGroup class

class CEnemyBattleGroup : public CDataEntry
{
//////////////////////////////////////////////////////////////////////
// Default implementation
public:

	enum class_info
	{
		ptr_address = 0x10C80D,			// Address of pointer table
		ptr_size = 8,					// Size of each entry in pointer table
		def_data_start = 0x10D74C,		// Location of original data
		def_data_extent = 0x10E1B4,		// Original extent of data
		num_entries = 484,
	};

	void WriteObject(FILE*, unsigned long address);
	void WritePtr(FILE*, unsigned long address);
	void ReadObject(FILE* file, int index, DWORD data = 0);

	int GetAddress();
	int GetSize();

	CEnemyBattleGroup();
	~CEnemyBattleGroup();

///////////////////////////////////////////////////////////////////////
// Class-specific data and behavior

	// Logical redundancies
	unsigned char entrycount;		// Number of entries in this group

	// Data stored in pointer table
	unsigned short runflag;			// Flag that toggles group behavior
	unsigned char flaguse;			// Use of the flag
	unsigned char letterbox;		// Battle letterbox size

	// Entry range data
	struct
	{
		unsigned char num;
		unsigned char enemy;
	} enemies[16];	// Only 16 entries allowed for sanity reasons

	unsigned short bg1;	// Background animation layers
	unsigned short bg2;

	// behavior
	bool AddEnemy(int num, int enemy);
	bool InsertEnemy(int pos, int num, int enemy);
	bool DeleteEnemy(int pos);
	int  MoveEnemy(int pos, int newpos);

	int CreateString(char* s, CRom* pRom);
};






//////////////////////////////////////////////////////////////////////
// CDoorEntry class

class CDoorEntry : public CDataEntry
{
//////////////////////////////////////////////////////////////////////
// Default implementation
public:

	enum class_info
	{
		ptr_address = 0x100200,			// Address of pointer table
		ptr_size = 4,					// Size of each entry in pointer table
		def_data_start = 0xF284F,		// Location of original data
		def_data_extent = 0xF5AE3,		// Original extent of data
		num_entries = 1280,				// Total number of entries of this type
	};

	void WriteObject(FILE*, unsigned long address);
	void WritePtr(FILE*, unsigned long address);
	void ReadObject(FILE* file, int index, DWORD data);

	int GetAddress();
	int GetSize();

///////////////////////////////////////////////////////////////////////
// Class-specific data and behavior

	CDoorEntry();

	
	CPtrArray doors;				// Array of pointers to door objects

	int num_doors;					// Logical redundancy

	void AddDoor(DoorEntry door);	// Add new door to array
	void RemDoor(int door);			// Remove specified door

	DoorEntry& operator[](int);

	// virtual destructor deletes all allocated doors
	~CDoorEntry();
};





//////////////////////////////////////////////////////////////////////
// CDataPack class

class CDataPack
{
public:
	CDataPack();

protected:
	// This class uses a CPtrArray to hold a list of pointers
	// to the data entries to be written
	CPtrArray m_data;

public:
	// Add a new pointer to a data entry
	void Add(CDataEntry *data);

	int GetCount();	// Returns number of objects in list
	int GetSize();	// Returns total size

	int GetStart();	// Returns first address
	int GetEnd();	// Returns address of end of data

	// Writes the package to the specified address
	void WritePack(FILE* file, unsigned long address);

	CDataEntry& operator[](int);

	~CDataPack();
};










//////////////////////////////////////////////////////////////////////
// CDestinationList class
// Maintains an array of door destination entries
// (Or rather, an array of pointers thereto)

// This class has no base; it can't really be derived from
// CDataEntry, because it represents an unorganized heap of
// data, rather than a neat list of entries along with a
// pointer table.





class CDestinationList
{

protected:
	CPtrArray m_data;

public:
	int m_nCount;

	void ReadData(FILE* file, unsigned long address, unsigned long typelist_address=NULL);

	// Writes the pile of destinations to the specified address.
	// Also writes a type-index list to the address specified by the third parameter.
	void WriteData(FILE* file, unsigned long address, unsigned long typelist_address);

	void AddDest(DestinationEntry d);	// Adds destination entry to array
	void RemDest(int n);				// Removes destination entry

	DestinationEntry& GetDest(int n);	// Returns a reference to a destination
	DestinationEntry& operator[](int);	// Array access operator

	CDestinationList();

	~CDestinationList();
};













//////////////////////////////////////////////////////////////////////
// CControlNode class

class CControlNode
{

public:
	int level;				// depth, i.e., CC byte # (sorta)
	short value;			// byte value (-1 for arguments)

	unsigned char type;		// code byte or argument
	unsigned char num_arg;	// number of argument bytes
	unsigned char args[32];	// argument list
	bool arg_multiplier;	// if true, number of arguments remaining is multiplied by the
							// value of the first

	bool isTerminator;

	int num_nodes;			// Number of child nodes


	CPtrArray *children;	// Array of pointers to child nodes

	// Add a new node, returning a pointer to the node created
	CControlNode* AddNode(int level, int value, int type, int num_arg, bool multi, bool terminate);


	
	void Free();			// Deletes child nodes before deleting self



};



